''********************************************
''*  Full-Duplex Serial Driver v1.2          *
''*  Author: Chip Gracey, Jeff Martin        *
''*  Copyright (c) 2006-2009 Parallax, Inc.  *
''*  See end of file for terms of use.       *
''********************************************

{-----------------REVISION HISTORY-----------------
2014nov22 na Versioning introduced, Copy Process Control Block from here to maine!
2014nov02 na FIXIT: da gibt es noch debug-mechanismen, die Sache muss nochmal berprft werden
2010feb23 na additional call of receiver in tx (->"1") to increase speed. Pointer in global mem slows down the speed
2010jan01 na LIN-driver modification: header needs about 2 bytes break signal (13 zero's)     CMap0221_FlowDiagram_fullduplex_serial
 v1.2 - 5/7/2009 fixed bug in dec method causing largest negative value (-2,147,483,648) to be output as -0.
 v1.1 - 3/1/2006 first official release.
}
CON
' Commands for SeCo
'  #1, Idle, Run, ParReLoa, LIN_Msg
  SecoVers = -4   ' Process version at return has to be =< 0 as the number of parameters passed at call is positive!
{
CON      NmbSeCoPar = 16 ' Number of pointers related to serial communication object     in PrCB Process Control Block
Dat PrCBSeCo long  @SeCoCmd,  @rxPinNmb, @txPinNmb, @rxBiTics, @txBiTics, @SeCoMode, @rxFreeee, @txFreeee
             long  @rxBuHead, @rxBuTail, @rxBuffer, @rxBuEnd_, @txBuHead, @txBuTail, @txBuffer, @txBuEnd_
DAT
SeCoCog  long 0         ' Cog running SerialCommunicationProcess
SeCoCmd  long 0         ' Command for serial communication
rxPinNmb long rxpin     ' pin number of input
txPinNmb long txpin     ' pin number of output
rxBiTics long 0         ' Number of counter tics per bit: 1/Baudrate
txBiTics long 0         ' Number of counter tics per bit: 1/Baudrate
SeCoMode long defMode   ' mode of serial connection
rxFreeee long 0         ' not used yet
txFreeee long 0         ' not used yet 
rxBuHead long 0         ' input buffer head: next input to get
rxBuTail long 0         ' input buffer tail: fetch received bytes from here
rxBuffer byte 0,0,0,0,0,0,0,0 ' Start address of input buffer
         byte 0,0,0,0,0,0,0
rxBuEnd_ byte 0         ' end address of input buffer
txBuHead long 0         ' output buffer head: next byte to be send         
txBuTail long 0         ' output buffer tail: place byte to send here
txBuffer byte 0,0,0,0,0,0,0,0 ' Start address of output buffer             
         byte 0,0,0,0,0,0,0
txBuEnd_ byte 0         ' end of output buffer                             
}
VAR

  long  cog                     'cog flag/id
{
  long  rx_head                 '9 contiguous longs
  long  rx_tail
  long  tx_head
  long  tx_tail
  long  rx_pin
  long  tx_pin
  long  rxtx_mode
  long  bit_ticks
  long  buffer_ptr
                     
  byte  rx_buffer[16]           'transmit and receive buffers
  byte  tx_buffer[16]  
}
CON
MaxCmd  = 5  'Numb                 ' Constant = number of allowed commands(functions) 
 
VAR
    byte ID_cog                    ' Global Variable number of the Cog, value is returned to calling process.

PUB Start( GlobPtr, AddrOffs ) | i                ' Start the object: store all the pointers to the global memory locally

  Stop  ' Start can restart, therefore stop a possibly running cog first
    ' Create a local copy of the global pointer array and do the address correction
    repeat i from 0 to LONG[LONG[GlobPtr +   0]+AddrOffs]-1        ' The first parameter is the number of parameters
      LONG[@LoclPtr + (i<<2)] := LONG[GlobPtr + (i<<2)] + AddrOffs ' all pointers are 32 bit, the pointer address is byte oriented                                  
    
  ifnot ID_cog
    ID_cog := cognew(@ASM_Entry, 0) + 1
  ' now wait for started process to return version #
  Repeat while ( i := LONG[LONG[GlobPtr +  0] + AddrOffs]) > 0
  Result := i<<4 + ID_Cog  '  Return 16* Version + ID-Cog
                
PRI Stop       ' Stop already running process
  if ID_cog
    cogstop(ID_cog~ - 1)
    
'***********************************
'* Assembly language serial driver *
'***********************************
DAT     org   0   ' data block defines assembly code block
        
ASM_Entry       ' Realize serial communication functions: simple full/duplex, ... LIN

        ' Initialize parameters and acknowlege cog launch command. 

        mov       Command,  #3          ' Fake "Reload Parameter" command at start up
        wrlong    Command,  CommandPtr  ' (Command -> *CommandPtr)  
  mov dira, DirAbits

        
LoopCmd
        rdlong    Command,  CommandPtr      ' Command queries (Command <- *CommandPtr) rdlong uses indirect addressing from main memory
        cmp       Command,  #MaxCmd       wc ' (Compare Command with #MaxCmd) C-bit is set if  Command < #MaxCmd
if_nc   mov       Command,  #0               ' If too big (Command >= MaxCmd), set C-bit it to zero
        add       Command,  #1            wz ' This Z provides for the functioning of LoopDre! (Command++) 
        add       Command,  #:JmpTab         ' (Command <- Command + #:JmpTab) address of the Jump table is #:JmpTab 
        movs      :JmpTab,  Command          ' (:JmpTab <- Command) ?
        nop                                  ' Nop because Prozessorpipeline delay
:JmpTab
        jmp       #0-0      ' Destination according to the following is Cmd  this jump is at assembly time jump to 0 during running it will take real address of Command  
        jmp       #LoopCmd  ' 0:   Wait for new command: idle mode
        jmp       #EndRun   ' 1:   Stop running processes, switch to Idle Mode
        jmp       #StrtRun  ' 2:   full duplex serial I/O like FullDuplexSerial from Parallax
        jmp       #ReLdPar  ' 3:   Parameters reload to reconfigure
        jmp       #LIN_Msg  ' 4:   Send outputbuffer as a LIN-Message
           
ReLdPar   ' Reload parameters to control communication process
          '
        rdlong    temp,     rxPinNmbPtr       ' pin number of receiver  to create mask
        mov       rxmask,   #1
        shl       rxmask,   temp              ' set bit corresponding to rx input
        rdlong    temp,     txPinNmbPtr       ' pin number of transmitter
        mov       txmask,   #1
        shl       txmask,   temp              ' set bit corresponding to tx output
        rdlong    rxBiTics, rxBiTicsPtr       ' Tics per bit: 1/baudrate
        rdlong    txBiTics, txBiTicsPtr       ' Tics per bit: 1/baudrate: option for asymmetric upload/download
        rdlong    rxtxMode, SeCoModePtr       ' mode control: polarity, local echo,..
        rdlong    rxFreeee, rxFreeeePtr       ' unused, for test items
        rdlong    txFreeee, txFreeeePtr       ' unused, for test items
        rdlong    rxBuffer, rxBufferPtr       ' Startaddress of rxBuffer
        rdlong    rxBuEnd_, rxBuEnd_Ptr       ' Endaddress of rxBuffer
        rdlong    txBuffer, txBufferPtr       ' Startaddress of txBuffer
        rdlong    txBuEnd_, txBuEnd_Ptr       ' Endaddress of txBuffer
        mov       rxBuHead, rxBufferPtr

        wrlong    txBufferPtr, txBuHeadPtr    ' next address to store a send-byte
        wrlong    txBufferPtr, txBuTailPtr    ' address last sended byte was read from
        wrlong    rxBufferPtr, rxBuHeadPtr    ' 
        wrlong    rxBufferPtr, rxBuTailPtr    ' 
        mov       LIN_Flag, #0          ' set flag false

' setup op-mode masks and output the stop bit already here:
        test    rxtxmode,#%100  wz    'tri-state output of TX pin
        test    rxtxmode,#%010  wc    'invert TX bit
        if_z_and_c      xor     txdata,   #1
                        shr     txdata,   #1    wc
        if_z            muxc    outa,txmask        
        if_nz           muxnc   dira,txmask




        mov       Command,  #0          ' aknowledge co
        wrlong    Command,  CommandPtr  ' Acknowledge: Process running
        jmp       #LoopCmd              ' wait for start command (alternatively: run fullduplex, could be controlled by mode parameter

EndRun
        mov       Command,  #0                ' (Command <- 0)
        wrlong    Command,  CommandPtr        ' Function acknowledge (Command -> *CommandPtr)
        jmp       #LoopCmd                    ' jump to address / label

LIN_Msg       ' Expand first byte to 16 Bit MARK as a header
              mov       LIN_Flag, #8          ' set flag true: also number of bits! tricky

StrtRun                                  
                        test    rxtxmode, #%100 wz    'init tx pin according to mode: open drain
                        test    rxtxmode, #%010 wc                                  ' invert
        if_z_ne_c       or      outa,     txmask      ' set tx pin to stop condition
        if_z            or      dira,     txmask      ' set tx pin to output if not open drain

                        mov     txcode,   #transmit   'initialize ping-pong multitasking to "call" transmit
        mov       Command,  #0                ' (Command <- 0)
        wrlong    Command,  CommandPtr        ' Function acknowledge (Command -> *CommandPtr)
                        rdlong  txBuTail, txBuTailPtr  ' initial read of tail
'
'
' Receive
'
receive                 jmpret  rxcode,   txcode      'run a chunk of transmit code, then return

                        test    rxtxmode, #%001 wz    ' zero when normal, nz when inverted
                        test    rxmask,   ina   wc    ' wait for start bit on rx pin to come        

        if_z_eq_c       jmp     #receive              ' loop when input == mode 

            OR      outa,  rxStrtSe   ' 

                        mov     rxbits,   #9          'ready to receive byte
                        mov     rxcnt,    rxBiTics
                        shr     rxcnt,    #2           ' middle. 2: corner
                        add     rxcnt,    cnt                      

            AND      outa,  rxStrtRe   ' 
:bit                    add     rxcnt,    rxBiTics    'ready next bit period

:wait                   jmpret  rxcode,txcode         'run a chuck of transmit code, then return
                        mov     temp,     rxcnt       'check if bit receive period done
                        sub     temp,     cnt
                        cmps    temp,     #0    wc
        if_nc           jmp     #:wait
            OR      outa,  rxToggSe   ' 
           jmpret  rxcode,   txcode      'run a chunk of transmit code, then return

                        test    rxmask,   ina   wc    'receive bit on rx pin
                        rcr     rxdata,   #1
             AND      outa,  rxToggRe   ' 
                        djnz    rxbits,   #:bit

                        shr     rxdata,   #32-9       'justify and trim received byte
                        and     rxdata,   #$FF
                        test    rxtxmode, #%001 wz    'if rx inverted, invert byte
        if_nz           xor     rxdata,   #$FF
                       
                        wrbyte  rxdata,   rxBuHead    ' write byte to global memory
           jmpret  rxcode,   txcode      'run a chunk of transmit code, then return
                        add     rxBuHead, #1
                        cmp     rxBuEnd_Ptr, rxBuHead  wc
        if_c            mov     rxBuHead, rxBufferPtr
           jmpret  rxcode,   txcode      'run a chunk of transmit code, then return
                        wrlong  rxBuHead, rxBuHeadPtr

                        jmp     #receive              'byte done, receive next byte
'
'
' Transmit
'
transmit
                        rdlong  Command,  CommandPtr ' Test on Idle-Command (Command <- *CommandPtr)
                        cmp     Command,  #1    wz ' compare Command with 1 (Command == 1 ?) if its equal Z-bit is set
        if_z            jmp #EndRun            ' if (Command == 1) go to :EndRun
         jmpret  txcode,   rxcode      'run a chunk of receive code, then return
                        rdlong  txBuHead, txBuHeadPtr 'check buffer empty ' 200 ns
                        cmp     txBuHead, txBuTail      wz                '  #

        if_z            jmp     #transmit             ' Tail == Head, empty   

            OR      outa,  txStrtSe   ' Show start of transmit   50ns
         jmpret  txcode,   rxcode      'run a chunk of receive code, then return
            AND      outa,  txStrtRe   ' Show start of transmit  50ns

                        rdbyte  txdata,   txBuTail    'get byte and inc tail
                        add     txBuTail, #1
                        cmp     txBuEnd_Ptr, txBuTail  wc
        if_c            mov     txBuTail, txBufferPtr
                        wrlong  txBuTail, txBuTailPtr
               jmpret  txcode,   rxcode      'run a chunk of receive code, then return    ->"1"

    
                  or      txdata, StopBits  '#$300          ' prepare byte to transmit two stopbits
                  shl     txdata,#1             ' shift data to create start bit
                  mov     txbits,#11            ' number of bits to send
                  shl     txdata, LIN_Flag      ' "enlarge" Start to wake up LIN slave
                  add     txbits, LIN_Flag            ' increase number of bits to send
                 mov     LIN_Flag, #0  ' set flag false

             ' invert data if tx invert mode
                    test    rxtxmode,#%100  wz    'output bit on tx pin according to mode
                                                  ' will be zero if P/P output
                    test    rxtxmode,#%010  wc    ' will be carry if invers polarity
                    
        if_z_and_c      xor     txdata,   invert      ' invert data 

 
               jmpret  txcode,   rxcode      'run a chunk of receive code, then return    ->"1"
                        mov     txcnt,cnt
txBit
                        add     txcnt,  txBiTics      'ready next cnt
                    test    rxtxmode,#%100  wz    'output bit on tx pin according to mode
                                                  ' will be zero if P/P output
                        shr     txdata,   #1    wc
        if_z            muxc    outa,txmask        
        if_nz           muxnc   dira,txmask

:wait
        '    OR      outa,  txToggSe   ' waitloop for send bit
                  jmpret  txcode,rxcode         'run a chunk of receive code, then return
                        mov     Temp,  txcnt          'check if bit transmit period done
                        sub     Temp,  cnt
                        cmps    Temp,  #0       wc

        '    AND      outa,  txToggRe   ' waitloop for send bit
        if_nc           jmp     #:wait

                        djnz    txbits,#txBit          'another bit to transmit?

                        jmp     #transmit             'byte done, transmit next byte
'
'
LoclPtr  ' All Pointers to Globalmemory: relates to ProcessControlBlock
CommandPtr  long 0       ' Command to control this process
rxPinNmbPtr long 0      ' pin number of input
txPinNmbPtr long 0      ' pin number of output
rxBiTicsPtr long 0      ' Number of counter tics per bit: 1/Baudrate
txBiTicsPtr long 0      ' Number of counter tics per bit: 1/Baudrate
SeCoModePtr long 0      ' mode of serial connection
rxFreeeePtr long 0      ' not used yet
txFreeeePtr long 0      ' not used yet 
rxBuHeadPtr long 0      ' input buffer head: next input to get
rxBuTailPtr long 0      ' input buffer tail: fetch received bytes from here
rxBufferPtr long 0      ' Start address of input buffer
rxBuEnd_Ptr long 0      ' end address of input buffer
txBuHeadPtr long 0      ' output buffer head: next byte to be send         
txBuTailPtr long 0      ' output buffer tail: place byte to send here
txBufferPtr long 0      ' Start address of output buffer                   
txBuEnd_Ptr long 0      ' end address of output buffer                     




Command  long 0          ' Command to control this process
rxMask   long 0         ' pin number of input as a mask
txMask   long 0         ' pin number of output
rxBiTics long 0         ' Number of counter tics per bit: 1/Baudrate
txBiTics long 0         ' Number of counter tics per bit: 1/Baudrate
rxtxMode long 0         ' mode of serial connection
rxFreeee long 0         ' not used yet
txFreeee long 0         ' not used yet 
rxBuHead long 0         ' input buffer head: next input to get
rxBuTail long 0         ' input buffer tail: fetch received bytes from here
rxBuffer long 0         ' Start address of input buffer
rxBuEnd_ long 0         ' end address of input buffer
txBuHead long 0         ' output buffer head: next byte to be send         
txBuTail long 0         ' output buffer tail: place byte to send here
txBuffer long 0         ' Start address of output buffer                   
txBuEnd_ long 0         ' end address of output buffer                     
LIN_Flag long 0         ' 1: create a long zero header

temp     long 0         ' temporary storage

'DirABits long %0000_1000_0000_1111_0000_0000_0000_0000      ' Bit 27: Counter A output (transmit)  and debug bits
DirABits long %0000_0000_0000_0000_0000_0000_0000_0000      ' FIXIT: hier gibt es noch debugging-mechanismen! kann wohl raus?
StopBits long %0000_0000_0000_0000_0000_0011_0000_0000                ' to insert stop bits
txStrtSe long %0000_0000_0000_0001_0000_0000_0000_0000
txToggSe long %0000_0000_0000_0010_0000_0000_0000_0000
rxStrtSe long %0000_0000_0000_0100_0000_0000_0000_0000
rxToggSe long %0000_0000_0000_1000_0000_0000_0000_0000

txStrtRe long %1111_1111_1111_1110_1111_1111_1111_1111
txToggRe long %1111_1111_1111_1101_1111_1111_1111_1111
rxStrtRe long %1111_1111_1111_1011_1111_1111_1111_1111
rxToggRe long %1111_1111_1111_0111_1111_1111_1111_1111
invert   long %1111_1111_1111_1111_1111_1111_1111_1111     ' also: -1
MSBit    long %1000_0000_0000_0000_0000_0000_0000_0000     ' to test the msb

rxbuff                  res     1
rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1         ' counter for receive bit duration
rxcode                  res     1         ' to store return address

txbuff                  res     1
txdata                  res     1
txbits                  res     1
txcnt                   res     1          ' counter for transmit bit duration
txcode                  res     1          ' to store return address

FIT